import pandas as pd
import wgs_analysis.algorithms.rearrangement
import scgenome.csvutils
import scgenome.loaders.utils
import scgenome.utils


def load_destruct_results(results_dir):
    """ Load destruct breakpoint data from a results directory.

    Args:
        results_dir (str): results directory
    """

    destruct_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'destruct_breakpoints.csv.gz', analysis_type='breakpoint_calling')

    return load_destruct_files(destruct_filepath)


def load_destruct_files(destruct_filepath):
    """ Read destruct breakpoint data from a file.

    Args:
        destruct_filepath (str): results filepath
    """

    csv_input = scgenome.csvutils.CsvInput(destruct_filepath)
    data = csv_input.read_csv()

    data['chromosome_1'] = data['chromosome_1'].astype(str)
    data['chromosome_2'] = data['chromosome_2'].astype(str)

    return data


def load_destruct_counts_results(results_dir):
    """ Load destruct breakpoint count data from a results directory.

    Args:
        results_dir (str): results directory
    """

    destruct_counts_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'destruct_cell_counts.csv.gz', analysis_type='breakpoint_calling')

    return load_destruct_counts_files(destruct_counts_filepath)


def load_destruct_counts_files(destruct_counts_filepath):
    """ Load destruct breakpoint count data from a file.

    Args:
        destruct_counts_filepath (str): results filepath
    """

    csv_input = scgenome.csvutils.CsvInput(destruct_counts_filepath)
    data = csv_input.read_csv()

    data = data.rename(columns={'cluster_id': 'prediction_id'})

    # KLUDGE: normal reads are not filtered properly, filter by their prefix, and having '-' in cell id
    data = data.loc[~data['cell_id'].str.startswith('HS'), :]
    data = data.loc[data['cell_id'].apply(lambda a: '-' in a), :]
    
    return data


def load_lumpy_results(results_dir, standardize_columns=False):
    """ Load lumpy breakpoint data from a results directory.

    Args:
        results_dir(str): results directory
        
    KwArgs:
        standardize_columns(bool): rename columns to be similar to destruct
    """

    lumpy_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'lumpy_breakpoints.csv.gz', analysis_type='breakpoint_calling')

    return load_lumpy_files(lumpy_filepath, standardize_columns=standardize_columns)


def load_lumpy_files(lumpy_filepath, standardize_columns=False):
    """ Load lumpy breakpoint data from a file.

    Args:
        lumpy_filepath(str): results filepath

    KwArgs:
        standardize_columns(bool): rename columns to be similar to destruct
    """

    csv_input = scgenome.csvutils.CsvInput(lumpy_filepath)
    data = csv_input.read_csv()

    count_cols = ['normal_PE', 'tumour_SR', 'normal_SR', 'tumour_PE']
    data[count_cols] = data[count_cols].fillna(0).astype(int)

    data = data[(data['normal_PE'] == 0) & (data['normal_SR'] == 0)]
    
    data['chrom1'] = data['chrom1'].astype(str)
    data['chrom2'] = data['chrom2'].astype(str)

    # Standardize columns
    if standardize_columns:
        data = data.rename(columns={
            'breakpoint_id': 'prediction_id',
            'chrom1': 'chromosome_1',
            'chrom2': 'chromosome_2',
            'strand1': 'strand_1',
            'strand2': 'strand_2',
        })

        type_translate = {
            'DELETION': 'deletion',
            'DUPLICATION': 'duplication',
            'INVERSION': 'inversion',
            'INTERCHROM': 'translocation',
        }
        data['type'] = data['type'].apply(lambda a: type_translate[a])

        # Note: its unclear whether lumpy uses 0-based or 1-based coordinates
        data['position_1'] = data[['confidence_interval_start1', 'confidence_interval_end1']].mean(axis=1)
        data['position_2'] = data[['confidence_interval_start2', 'confidence_interval_end2']].mean(axis=1)

        data['position_1'] = data['position_1'].astype(int)
        data['position_2'] = data['position_2'].astype(int)

    return data


def load_lumpy_counts_results(results_dir):
    """ Load lumpy breakpoint counts data from a results directory.

    Args:
        results_dir (str): results directory
    """

    lumpy_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'lumpy_breakpoints_evidence.csv.gz', analysis_type='breakpoint_calling')

    return load_lumpy_counts_files(lumpy_filepath)


def load_lumpy_counts_files(lumpy_filepath):
    """ Load lumpy breakpoint counts data from a file.

    Args:
        lumpy_filepath (str): results filepath
    """

    csv_input = scgenome.csvutils.CsvInput(lumpy_filepath)
    data = csv_input.read_csv()

    # KLUDGE: normal reads are not filtered properly, filter by their prefix, and having '-' in cell id
    data = data.loc[~data['cell_id'].str.startswith('HS'), :]
    data = data.loc[data['cell_id'].apply(lambda a: '-' in a), :]

    return data


def load_consensus_breakpoint_results(results_dir):
    """ Load / calculate consensus breakpoints from a results directory.

    Args:
        results_dir (str): results directory
    """

    destruct_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'destruct_breakpoints.csv.gz', analysis_type='breakpoint_calling')

    lumpy_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'lumpy_breakpoints.csv.gz', analysis_type='breakpoint_calling')

    destruct_counts_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'destruct_cell_counts.csv.gz', analysis_type='breakpoint_calling')

    return load_consensus_breakpoint_files(destruct_filepath, lumpy_filepath, destruct_counts_filepath)


def load_consensus_breakpoint_files(destruct_filepath, lumpy_filepath, destruct_counts_filepath):
    """ Load / calculate consensus breakpoints from files.

    Args:
        results_dir (str): results directory
    """

    destruct_breakpoints = scgenome.loaders.breakpoint.load_destruct_files(destruct_filepath)
    lumpy_breakpoints = scgenome.loaders.breakpoint.load_lumpy_files(lumpy_filepath, standardize_columns=True)

    destruct_breakpoint_counts = scgenome.loaders.breakpoint.load_destruct_counts_files(destruct_counts_filepath)

    destruct_lumpy_matches = wgs_analysis.algorithms.rearrangement.match_breakpoints(
        lumpy_breakpoints, destruct_breakpoints, window_size=200)

    matched_destruct_predictions = (
        destruct_lumpy_matches[['target_id']].drop_duplicates()
        .rename(columns={'target_id': 'prediction_id'}))

    # Filter destruct by lumpy matches
    destruct_breakpoints = destruct_breakpoints.merge(matched_destruct_predictions)
    destruct_breakpoint_counts = destruct_breakpoint_counts.merge(matched_destruct_predictions)

    return {
        'breakpoint_data': destruct_breakpoints,
        'breakpoint_count_data': destruct_breakpoint_counts,
    }




import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from anndata import AnnData

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch
import matplotlib.cm

import scgenome.cnplot
import scgenome.refgenome


# TODO: refactor color functionality elsewhere
color_reference = {
    0:'#3182BD',
    1:'#9ECAE1',
    2:'#CCCCCC',
    3:'#FDCC8A',
    4:'#FC8D59',
    5:'#E34A33',
    6:'#B30000',
    7:'#980043',
    8:'#DD1C77',
    9:'#DF65B0',
    10:'#C994C7',
    11:'#D4B9DA',
}


def hex_to_rgb(h):
    if h is None:
        return np.array((0, 0, 0), dtype=int)
    h = h.lstrip('#')
    return np.array(tuple(np.uint8(int(h[i:i+2], 16)) for i in (0, 2 ,4)), dtype=int)


def plot_cell_cn_matrix(
        adata: AnnData,
        layer_name='state',
        cell_order_fields=(),
        ax=None,
        ax_legend=None,
        raw=False,
        max_cn=13,
        show_cell_ids=False):
    """ Plot a copy number matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    cell_order_fields : list, optional
        columns of obs on which to sort cells, by default None
    ax : matplotlib.axes.Axes, optional
        existing axis to plot into, by default None
    raw : bool, optional
        raw plotting, no integer color map, by default False
    max_cn : int, optional
        clip cn at max value, by default 13
    show_cell_ids : bool, optional
        show cell ids on heatmap axis, by default False

    Returns
    -------
    dict
        Dictionary of plot and data elements

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()
        scgenome.pl.plot_cell_cn_matrix(adata)

    """

    if ax is None:
        ax = plt.gca()

    # Order the chromosomes
    chr_start = adata.var.reset_index().merge(scgenome.refgenome.info.chromosome_info[['chr', 'chr_index']], how='left')
    if chr_start['chr_index'].isnull().any():
        chromosomes = adata.var['chr'].astype(str).values
        raise ValueError(f'mismatching chromosomes {chromosomes} and {scgenome.refgenome.info.chromosomes}')
    chr_start = chr_start[['start', 'chr_index']].values
    genome_ordering = np.lexsort(chr_start.transpose())

    # Order the cells if requested
    if len(cell_order_fields) > 0:
        cell_order_fields = reversed(list(cell_order_fields))
        cell_order_values = adata.obs[cell_order_fields].values.transpose()
        cell_ordering = np.lexsort(cell_order_values)

    else:
        cell_ordering = range(adata.shape[0])

    adata = adata[cell_ordering, genome_ordering]

    if layer_name is not None:
        X = adata.layers[layer_name].copy()
    else:
        X = adata.X.copy()

    X = np.nan_to_num(X, nan=0)

    if max_cn is not None:
        X[X > max_cn] = max_cn

    cmap = None
    if not raw:
        X_colors = np.zeros(X.shape + (3,), dtype=int)
        X_colors[X < 0, :] = 0
        X_colors[X > max(color_reference.keys()), :] = max(color_reference.keys())
        for state, hex in color_reference.items():
            X_colors[X == state, :] = hex_to_rgb(hex)
        im = ax.imshow(X_colors, aspect='auto', cmap=cmap, interpolation='none')

    else:
        im = sns.heatmap(X, ax=ax, cbar=False)

    mat_chrom_idxs = chr_start[genome_ordering][:, 1]
    chrom_boundaries = np.array([0] + list(np.where(mat_chrom_idxs[1:] != mat_chrom_idxs[:-1])[0]) + [mat_chrom_idxs.shape[0] - 1])
    chrom_sizes = chrom_boundaries[1:] - chrom_boundaries[:-1]
    chrom_mids = chrom_boundaries[:-1] + chrom_sizes / 2
    ordered_mat_chrom_idxs = mat_chrom_idxs[np.where(np.array([1] + list(np.diff(mat_chrom_idxs))) != 0)]
    chrom_names = np.array(scgenome.refgenome.info.plot_chromosomes)[ordered_mat_chrom_idxs]

    ax.set_xticks(chrom_mids)
    ax.set_xticklabels(chrom_names, fontsize='6')

    if show_cell_ids:
        ax.set(yticks=range(len(adata.obs.index)))
        ax.set(yticklabels=adata.obs.index.values)
    else:
        ax.set(yticklabels=[])

    for val in chrom_boundaries[:-1]:
        ax.axvline(x=val, linewidth=1, color='black', zorder=100)

    legend = None
    if ax_legend is not None:
        states = []
        patches = []
        for s, h in color_reference.items():
            states.append(s)
            patches.append(Patch(facecolor=h, edgecolor=h))
        legend = ax_legend.legend(patches, states, ncol=3,
            frameon=True, loc=2, bbox_to_anchor=(0., 1.),
            facecolor='white', edgecolor='white', fontsize='4',
            title='Copy Number', title_fontsize='6')

    return {
        'ax': ax,
        'ax_legend': ax_legend,
        'im': im,
        'legend': legend,
        'adata': adata,
    }


def map_catagorigal_colors(values, cmap_name=None):
    levels = np.unique(values)
    n_levels = len(levels)

    if cmap_name is None:
        if n_levels <= 10:
            cmap_name = 'tab10'
        elif n_levels <= 20:
            cmap_name = 'tab20'
        else:
            cmap_name = 'hsv'

    cmap = matplotlib.cm.get_cmap(cmap_name)

    level_colors = dict(zip(levels, cmap(np.linspace(0, 1, n_levels))))

    value_colors = np.zeros(values.shape + (4,))
    for l, c in level_colors.items():
        value_colors[values == l, :] = c

    return level_colors, value_colors


def _plot_categorical_annotation(values, ax, ax_legend, title):
        level_colors, value_colors = map_catagorigal_colors(values)

        im = ax.imshow(value_colors, aspect='auto', interpolation='none')

        levels = []
        patches = []
        for s, h in level_colors.items():
            levels.append(s)
            patches.append(Patch(facecolor=h, edgecolor=h))
        legend = ax_legend.legend(patches, levels, ncol=3,
            frameon=True, loc=2, bbox_to_anchor=(0., 1.),
            facecolor='white', edgecolor='white', fontsize='4',
            title=title, title_fontsize='6')

        ax.grid(False)
        ax.set_xticks([0.], [title], rotation=90, fontsize='6')
        ax.set_yticks([])

        annotation_info = {}
        annotation_info = {}
        annotation_info['ax'] = ax
        annotation_info['im'] = im
        annotation_info['ax_legend'] = ax_legend
        annotation_info['legend'] = legend

        return annotation_info


def _plot_continuous_annotation(values, ax, ax_legend, title):

    im = ax.imshow(values, aspect='auto', interpolation='none', cmap='Reds')

    ax.grid(False)
    ax.set_xticks([0.], [title], rotation=90, fontsize='6')
    ax.set_yticks([])

    ax_legend.grid(False)
    ax_legend.set_xticks([])
    ax_legend.set_yticks([])

    axins = ax_legend.inset_axes([0.5, 0.1, 0.05, 0.6])

    cbar = plt.colorbar(im, cax=axins)
    axins.set_title(title, fontsize='6')
    cbar.ax.tick_params(labelsize='4')

    return {}


def plot_cell_cn_matrix_clusters_fig(
        adata: AnnData,
        layer_name='state',
        cell_order_fields=(),
        annotation_fields=(),
        fig=None,
        raw=False,
        max_cn=13,
        show_cell_ids=False):
    """ Plot a copy number matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    cell_order_fields : list, optional
        columns of obs on which to sort cells, by default None
    annotation_fields : list, optional
        column of obs to use as an annotation colorbar, by default 'cluster_id'
    fig : matplotlib.figure.Figure, optional
        existing figure to plot into, by default None
    raw : bool, optional
        raw plotting, no integer color map, by default False
    max_cn : int, optional
        clip cn at max value, by default 13
    show_cell_ids : bool, optional
        show cell ids on heatmap axis, by default False

    Returns
    -------
    dict
        Dictionary of plot and data elements

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()

        g = scgenome.pl.plot_cell_cn_matrix_clusters_fig(
            adata,
            cell_order_fields=['cell_order'],
            annotation_fields=['cluster_id', 'sample_id', 'quality'])

    """

    if fig is None:
        fig = plt.figure()

    fig_main, fig_legends = fig.subfigures(nrows=2, ncols=1, height_ratios=[5, 1], squeeze=True)
    fig_legends.patch.set_alpha(0.0)

    width_ratios = [1] + [0.005] + [0.02] * len(annotation_fields)

    axes = fig_main.subplots(
        nrows=1, ncols=len(width_ratios), sharey=True, width_ratios=width_ratios,
        squeeze=False, gridspec_kw=dict(hspace=0.02, wspace=0.02))[0]
    axes[1].set_axis_off()

    axes_legends = fig_legends.subplots(
        nrows=1, ncols=1+len(annotation_fields), squeeze=False)[0]
    for ax in axes_legends:
        ax.set_axis_off()
        ax.set_alpha(0.0)
        ax.patch.set_alpha(0.0)

    ax = axes[0]
    ax_legend = axes_legends[0]
    g = plot_cell_cn_matrix(
        adata, layer_name=layer_name,
        cell_order_fields=cell_order_fields,
        ax=ax, ax_legend=ax_legend, raw=raw, max_cn=max_cn,
        show_cell_ids=show_cell_ids)

    adata = g['adata']

    annotation_info = {}

    for ax, ax_legend, annotation_field in zip(axes[2:], axes_legends[1:], annotation_fields):
        if adata.obs[annotation_field].dtype.name in ('category', 'object'):
            values = adata.obs[[annotation_field]].values
            annotation_info[annotation_field] = _plot_categorical_annotation(values, ax, ax_legend, annotation_field)

        else:
            values = adata.obs[[annotation_field]].values
            annotation_info[annotation_field] = _plot_continuous_annotation(values, ax, ax_legend, annotation_field)

    return {
        'fig': fig,
        'axes': axes,
        'adata': adata,
        'annotation_info': annotation_info,
    }



import os
import logging
import seaborn
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

import wgs_analysis.snvs.mutsig
import wgs_analysis.plots.snv
import wgs_analysis.annotation.position


def filter_snv_data(
        snv_data,
        snv_count_data,
        num_cells_threshold=None,
        sum_alt_threshold=None,
        figures_prefix=None,
    ):
    """ Load filtered SNV annotation and count data

    Args:
        snv_data (pandas.DataFrame): snv annotation data table
        snv_count_data (pandas.DataFrame): snv count data table
        num_cells_threshold (int, optional): minimum number of cells threshold. Defaults to None.
        sum_alt_threshold (int, optional): minimum total alt count threshold. Defaults to None.
        figures_prefix (str, optional): filename prefix for figures. Defaults to None.

    Returns:
        pandas.DataFrame, pandas.DataFrame: SNV annotations, SNV counts
    """
    logging.info('Filtering and annotating SNVs')

    # Calculate cell counts
    cell_counts = (
        snv_count_data
        .query('alt_counts > 0')
        .drop_duplicates(['chrom', 'coord', 'ref', 'alt', 'cell_id'])
        .groupby(['chrom', 'coord', 'ref', 'alt'], observed=True).size().rename('num_cells')
        .reset_index())

    fig = plt.figure(figsize=(10, 4))
    cell_counts['num_cells'].astype(float).hist(bins=50)
    plt.xlabel('Num. Cells')
    plt.ylabel('Log Num SNVs')
    plt.gca().set_yscale('log')
    if figures_prefix is not None:
        fig.savefig(figures_prefix + 'snv_cell_counts.pdf', bbox_inches='tight')

    snv_data = snv_data.merge(cell_counts, how='left')
    if snv_data['num_cells'].isnull().any():
        num_no_count_snvs = snv_data['num_cells'].isnull().sum()
        logging.warning(f'found {num_no_count_snvs} with no counts in genotyping results')
        snv_data['num_cells'] = snv_data['num_cells'].fillna(0).astype(int)

    # Calculate total alt counts for each SNV
    sum_alt_counts = (
        snv_count_data
        .groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['alt_counts']
        .sum().rename('sum_alt_counts')
        .reset_index())

    fig = plt.figure(figsize=(10, 4))
    sum_alt_counts['sum_alt_counts'].astype(float).hist(bins=50)
    plt.xlabel('Total Alt. Counts')
    plt.ylabel('Log Num SNVs')
    plt.gca().set_yscale('log')
    if figures_prefix is not None:
        fig.savefig(figures_prefix + 'snv_alt_counts.pdf', bbox_inches='tight')

    # Filter SNVs by num cells in which they are detected
    if num_cells_threshold is not None:
        filtered_cell_counts = cell_counts.query(
            'num_cells >= {}'.format(num_cells_threshold))
        logging.info('Filtering {} of {} SNVs by num_cells >= {}'.format(
            len(filtered_cell_counts.index), len(cell_counts.index), num_cells_threshold))
        snv_data = snv_data.merge(filtered_cell_counts[['chrom', 'coord', 'ref', 'alt']].drop_duplicates())

    # Filter SNVs by total alt counts
    if sum_alt_threshold is not None:
        filtered_sum_alt_counts = sum_alt_counts.query(
            'sum_alt_counts >= {}'.format(sum_alt_threshold))
        logging.info('Filtering {} of {} SNVs by sum_alt_counts >= {}'.format(
            len(filtered_sum_alt_counts.index), len(sum_alt_counts.index), sum_alt_threshold))
        snv_data = snv_data.merge(filtered_sum_alt_counts[['chrom', 'coord', 'ref', 'alt']].drop_duplicates())

    snv_count_data = snv_count_data.merge(snv_data[['chrom', 'coord', 'ref', 'alt']].drop_duplicates())

    return snv_data, snv_count_data


def plot_mutation_signatures(snv_data, snv_class_col, figures_prefix=None):
    """
    Infer mutation signature probabilities and plot.
    """

    sigs, sig_prob = wgs_analysis.snvs.mutsig.load_signature_probabilities()

    snv_data = snv_data[snv_data['tri_nucleotide_context'].notnull()]
    snv_data = snv_data.merge(sigs)

    # Add in an all snvs category
    snv_data_all = snv_data.copy()
    snv_data_all[snv_class_col] = 'All'
    snv_data = pd.concat([snv_data, snv_data_all])

    # Generate signature distributions for cell count classes
    sample_sig = wgs_analysis.snvs.mutsig.fit_sample_signatures(
        snv_data.drop_duplicates(
            ['chrom', 'coord', 'ref', 'alt', snv_class_col]),
        sig_prob, snv_class_col)

    fig = wgs_analysis.snvs.mutsig.plot_signature_heatmap(sample_sig)
    seaborn.despine(trim=True)
    if figures_prefix is not None:
        fig.savefig(figures_prefix + f'_{snv_class_col}_mutsig.pdf', bbox_inches='tight')

    fig = plt.figure(figsize=(4, 4))
    if len(sample_sig.index) > 0:
        sample_sig.loc['All', :].sort_values().iloc[-10:,].plot(kind='bar')
    seaborn.despine(trim=True)
    if figures_prefix is not None:
        fig.savefig(figures_prefix + f'_{snv_class_col}_top10_mutsig.pdf', bbox_inches='tight')


def run_mutation_signature_analysis(snv_data, figures_prefix=None):
    """
    Run a mutation signature analysis
    """

    # Per cell count class signatures
    snv_data = snv_data[snv_data['num_cells'] > 0]
    snv_data['num_cells_class'] = '1'
    snv_data.loc[snv_data['num_cells'] > 1, 'num_cells_class'] = '2-5'
    snv_data.loc[snv_data['num_cells'] > 5, 'num_cells_class'] = '6-20'
    snv_data.loc[snv_data['num_cells'] > 20, 'num_cells_class'] = '>20'

    plot_mutation_signatures(snv_data, 'num_cells_class', figures_prefix=figures_prefix)

    # Adjacent distance class signatures
    snv_data['adjacent_distance_class'] = 'standard'
    snv_data.loc[snv_data['adjacent_distance'] <= 10000, 'adjacent_distance_class'] = 'hypermutated'

    plot_mutation_signatures(snv_data, 'adjacent_distance_class', figures_prefix=figures_prefix)


def run_bulk_snv_analysis(snv_data, snv_count_data, filtered_cell_ids, results_prefix=None):
    # Filter cells
    snv_count_data = snv_count_data.merge(filtered_cell_ids)
    total_alt_counts = snv_count_data.groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['alt_counts'].sum().reset_index()
    snv_data = snv_data.merge(
        total_alt_counts.query('alt_counts > 0')[['chrom', 'coord', 'ref', 'alt']].drop_duplicates())

    # Write high impact SNVs to a csv table
    high_impact = snv_data.query('effect_impact == "HIGH"')
    high_impact = high_impact.merge(
        snv_count_data.groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['alt_counts'].sum().reset_index())
    high_impact = high_impact.merge(
        snv_count_data.groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['ref_counts'].sum().reset_index())
    high_impact = high_impact.merge(
        snv_count_data.groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['total_counts'].sum().reset_index())
    if results_prefix is not None:
        high_impact.to_csv(results_prefix + 'snvs_high_impact.csv')

    # Annotate adjacent distance
    snv_data = wgs_analysis.annotation.position.annotate_adjacent_distance(snv_data)

    # Plot adjacent distance of SNVs
    fig = plt.figure(figsize=(10, 3))
    wgs_analysis.plots.snv.snv_adjacent_distance_plot(plt.gca(), snv_data)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'snv_adjacent_distance.pdf', bbox_inches='tight')

    # Plot snv count as a histogram across the genome
    fig = plt.figure(figsize=(10, 3))
    wgs_analysis.plots.snv.snv_count_plot(plt.gca(), snv_data)
    if results_prefix is not None:
        fig.savefig(results_prefix + 'snv_genome_count.pdf', bbox_inches='tight')

    # Run mutation signature analysis, requires adjacent distance
    run_mutation_signature_analysis(snv_data, results_prefix)



import anndata as ad
import pandas as pd

from typing import List
from anndata import AnnData


def ad_concat_cells(adatas: List[AnnData]) -> AnnData:
    """ Concatenate a list of anndata by obs (cells)

    Parameters
    ----------
    adatas : List[AnnData]
        list of anndata to concatenate

    Returns
    -------
    AnnData
        concatenated anndata with all obs from the input anndatas
    """    

    var_temp = pd.concat((a.var for a in adatas)).drop_duplicates()
    adata = ad.concat(adatas, join='outer')
    adata.var = var_temp.reindex(adata.var.index)

    return adata



import logging
import sklearn.cluster
import umap
import pandas as pd
import numpy as np
import anndata as ad
from natsort import natsorted

from anndata import AnnData
from typing import Dict, Any

import scgenome.cncluster
import scgenome.preprocessing.transform


def cluster_cells(adata: AnnData, layer_name='copy', method='kmeans') -> AnnData:
    """ Cluster cells by copy number.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    method : str, optional
        clustering method, by default 'kmeans'

    Returns
    -------
    AnnData
        copy number data with additional `cluster_id` column
    """

    if method == 'kmeans':
        cluster_cells_kmeans(adata, layer_name=layer_name)


def cluster_cells_kmeans(adata: AnnData, layer_name='copy', min_k=2, max_k=100) -> AnnData:
    """ Cluster cells by copy number using kmeans.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default 'state'
    min_k : int, optional
        minimum number of clusters, by default 2
    max_k : int, optional
        maximum number of clusters, by default 100

    Returns
    -------
    AnnData
        copy number data with additional `cluster_id` column

    Examples
    -------

    >>> import scgenome
    >>> import anndata as ad
    >>> import numpy as np
    >>> adata = ad.AnnData(np.array([
    ...    [3, 3, 3, 6, 6],
    ...    [1, 1, 1, 2, 2],
    ...    [1, 22, 1, 2, 2],
    ...    [1, 3, 3, 5, 5],
    ... ]).astype(np.float32))
    >>> adata = scgenome.tl.cluster_cells_kmeans(adata, layer_name=None, max_k=3)
    >>> adata.obs['cluster_id']
    0    0
    1    2
    2    1
    3    0
    Name: cluster_id, dtype: category
    Categories (3, int64): [0, 1, 2]

    """

    if layer_name is not None:
        X = adata.layers[layer_name]
    else:
        X = adata.X

    X = scgenome.preprocessing.transform.fill_missing(X)

    ks = range(min_k, max_k + 1)

    logging.info(f'trying with max k={max_k}')

    kmeans = []
    bics = []
    for k in ks:
        logging.info(f'trying with k={k}')
        model = sklearn.cluster.KMeans(n_clusters=k, init='k-means++', random_state=100).fit(X)
        bic = scgenome.cncluster.compute_bic(model, X)
        kmeans.append(model)
        bics.append(bic)

    opt_k = np.array(bics).argmax()
    logging.info(f'selected k={opt_k}')

    model = kmeans[opt_k]

    adata.obs['cluster_id'] = pd.Series(model.labels_, index=adata.obs.index, dtype='category')

    # store information on the clustering parameters
    adata.uns['kmeans'] = {}
    adata.uns['kmeans']['params'] = dict(
        opt_k=opt_k,
        min_k=min_k,
        max_k=max_k,
        layer_name=layer_name,
    )

    return adata


def aggregate_clusters(
        adata: AnnData,
        agg_X: Any,
        agg_layers: Dict=None,
        agg_obs: Dict=None,
        cluster_col: str='cluster_id',
        cluster_size_col: str='cluster_size') -> AnnData:
    """ Aggregate copy number by cluster to create cluster CN matrix

    Parameters
    ----------
    adata : AnnData
        copy number data
    agg_X : Any
        function to aggregate X
    agg_layers : Dict, optional
        functions to aggregate layers keyed by layer names, by default None
    agg_obs : Dict, optional
        functions to aggregate obs data keyed by obs columns, by default None
    cluster_col : str, optional
        column with cluster ids, by default 'cluster_id'
    cluster_size_col : str, optional
        column that will be set to the size of each cluster, by default 'cluster_size'

    Returns
    -------
    AnnData
        aggregated cluster copy number
    """

    X = (
        adata
            .to_df()
            .set_index(adata.obs[cluster_col].astype(str))
            .groupby(level=0)
            .agg(agg_X)
            .sort_index())

    layer_data = None
    if agg_layers is not None:
        layer_data = {}
        for layer_name in agg_layers:
            layer_data[layer_name] = (
                adata
                    .to_df(layer=layer_name)
                    .set_index(adata.obs[cluster_col].astype(str))
                    .groupby(level=0)
                    .agg(agg_layers[layer_name])
                    .sort_index())

    obs_data = {}
    obs_data[cluster_size_col] = (
        adata.obs
            .set_index(adata.obs[cluster_col].astype(str))
            .groupby(level=0)
            .size())

    if agg_obs is not None:
        for obs_name in agg_obs:
            obs_data[obs_name] = (
                adata.obs
                    .set_index(adata.obs[cluster_col].astype(str))[obs_name]
                    .groupby(level=0)
                    .agg(agg_obs[obs_name])
                    .sort_index())

    obs_data = pd.DataFrame(obs_data)

    adata = ad.AnnData(
        X,
        obs=obs_data,
        var=adata.var,
        layers=layer_data,
    )

    return adata


def aggregate_clusters_hmmcopy(adata: AnnData) -> AnnData:
    """ Aggregate hmmcopy copy number by cluster to create cluster CN matrix

    Parameters
    ----------
    adata : AnnData
        hmmcopy copy number data

    Returns
    -------
    AnnData
        aggregsated cluster copy number
    """

    agg_X = np.sum

    agg_layers = {
        'copy': np.nanmean,
        'state': np.nanmedian,
    }

    agg_obs = {
        'total_reads': np.nansum,
    }

    return aggregate_clusters(adata, agg_X, agg_layers, agg_obs, cluster_col='cluster_id')


def compute_umap(
        adata: AnnData,
        layer_name: str='copy',
        n_components: int=2,
        n_neighbors: int=15,
        min_dist: float=0.1,
        metric: str='euclidean',
    ) -> AnnData:
    """ Cluster cells by copy number.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data on which to perform umap dimensionality
        reduction, None for X, by default 'copy'
    n_components : int
        umap n_components param
    n_neighbors : int
        umap n_neighbors param
    min_dist : float
        umap min_dist param

    Returns
    -------
    AnnData
        copy number data with additional `umap_1`, `umap_2` columns
    """

    if layer_name is not None:
        X = adata.layers[layer_name]
    else:
        X = adata.X

    X = scgenome.preprocessing.transform.fill_missing(X)

    embedding = umap.UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        n_components=n_components,
        metric=metric,
        random_state=42,
    ).fit_transform(X)

    adata.obs['UMAP1'] = embedding[:, 0]
    adata.obs['UMAP2'] = embedding[:, 1]

    return adata



import pyfaidx
import pyBigWig
import numba
import pyranges as pr
import numpy as np

import scgenome.refgenome


def create_bins(binsize):
    """ Create a regular binning of the genome

    Parameters
    ----------
    binsize : int
        length of bins

    Returns
    -------
    pyrange.PyRanges
        regular bins tiled across the genome
    """    
    
    chromsizes = scgenome.refgenome.info.chromosome_info[['chr', 'chromosome_length']].rename(
        columns={'chr': 'Chromosome', 'chromosome_length': 'End'}).assign(Start=0)[['Chromosome', 'Start', 'End']]

    chromsizes = pr.PyRanges(chromsizes)

    bins = pr.gf.tile_genome(chromsizes, binsize)

    return bins


def _chromosome_count_gc(data, **kwargs):
    genome_fasta = kwargs['genome_fasta']
    chromosome = kwargs['chromosome']
    column_name = kwargs['column_name']

    genome = pyfaidx.Fasta(genome_fasta)
    chrom_sequence = np.fromiter(str(genome[chromosome]).upper(), dtype='<U1')

    gc_indicator = ((chrom_sequence == 'G') | (chrom_sequence == 'C')) * 1
    gc_cumsum = gc_indicator.cumsum()

    data[column_name] = gc_cumsum[data['End'].values - 1] - gc_cumsum[data['Start'].values]

    return data


def count_gc(bins, genome_fasta, column_name='gc', proportion=False):
    """ Count gc in each bin

    Parameters
    ----------
    bins : pyranges.PyRanges
        ranges for which to count gc
    genome_fasta : str
        reference genome fasta
    column_name : str, optional
        column to add to `bins`, by default 'gc'
    proportion : bool, optional
        proportion of length, by default False

    Returns
    -------
    pyranges.PyRanges
        output ranges with additional column for gc count
    """    

    data = bins.apply(_chromosome_count_gc, genome_fasta=genome_fasta, column_name=column_name)

    if proportion:
        data = data.assign(column_name, lambda df: df[column_name] / (df['End'] - df['Start']))

    return data


@numba.jit(nopython=True)
def mean_ranges(values, sequence, starts, ends):
    for i in range(starts.shape[0]):
        values[i] = np.nanmean(sequence[starts[i]:ends[i]])


def _chromosome_mean_bigwig(data, **kwargs):
    bigwig_file = kwargs['bigwig_file']
    chromosome = kwargs['chromosome']
    column_name = kwargs['column_name']
    chr_prefix = kwargs['chr_prefix']

    chromosome = chr_prefix + chromosome
    
    bw = pyBigWig.open(bigwig_file, 'r')

    if chromosome not in bw.chroms():
        data[column_name] = None
        return data

    chromosome_length = bw.chroms()[chromosome]
    sequence = np.array(bw.values(chromosome, start=0, end=chromosome_length))

    x = np.zeros(data.shape[0])
    mean_ranges(x, sequence, data['Start'].values, data['End'].values)

    data[column_name] = x

    return data


def mean_from_bigwig(bins, bigwig_file, column_name, chr_prefix=''):
    """ Count gc in each bin

    Parameters
    ----------
    bins : pyranges.PyRanges
        ranges for which to count gc
    bigwig_file : str
        bigwig filename
    column_name : str
        column to add to `bins`
    chr_prefix : str
        prefix for chromosome names, default ''

    Returns
    -------
    pyranges.PyRanges
        output ranges with additional column for mean bigwig value per bin
    """    

    data = bins.apply(
        _chromosome_mean_bigwig,
        bigwig_file=bigwig_file,
        column_name=column_name,
        chr_prefix=chr_prefix,
    )

    return data






import anndata as ad
import pkg_resources

from anndata import AnnData


def OV2295_HMMCopy_reduced() -> AnnData:
    """ DLP data from the OV2295 ovarian cell lines.

    Returns
    -------
    AnnData
        HMMCopy data, reduced size
    """

    adata_filename = pkg_resources.resource_filename('scgenome', 'datasets/data/OV2295_HMMCopy_reduced.h5ad')
    return ad.read(adata_filename)



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from anndata import AnnData

import scgenome.cnplot


def plot_cn_profile(
        adata: AnnData,
        obs_id: str,
        value_layer_name=None,
        state_layer_name=None,
        ax=None,
        max_cn=13,
        chromosome=None,
        s=5,
        squashy=False,
        rawy=False,
    ):
    """Plot scatter points of copy number across the genome or a chromosome.

    Parameters
    ----------
    adata : AnnData
        copy number data
    obs_id : str
        observation to plot
    value_layer_name : str, optional
        layer with values for y axis, None for X, by default None
    state_layer_name : str, optional
        layer with states for colors, None for no color by state, by default None
    ax : [type], optional
        existing axis to plot into, by default None
    max_cn : int, optional
        max copy number for y axis, by default 13
    chromosome : [type], optional
        single chromosome plot, by default None
    s : int, optional
        size of scatter points, by default 5
    squashy : bool, optional
        compress y axis, by default False
    rawy : bool, optional
        raw data on y axis, by default False

    Examples
    -------

    .. plot::
        :context: close-figs

        import scgenome
        adata = scgenome.datasets.OV2295_HMMCopy_reduced()
        scgenome.pl.plot_cn_profile(adata, 'SA922-A90554B-R27-C43', value_layer_name='copy', state_layer_name='state')

    TODO: missing return
    """

    cn_data = adata.var.copy()

    if value_layer_name is not None:
        cn_data['value'] = np.array(adata[[obs_id], :].layers[value_layer_name][0])
    else:
        cn_data['value'] = np.array(adata[[obs_id], :].X[0])

    cn_field_name = None
    if state_layer_name is not None:
        cn_data['state'] = np.array(adata[[obs_id], :].layers[state_layer_name][0])
        cn_field_name = 'state'

    if ax is None:
        ax = plt.gca()

    cn_data = cn_data.dropna(subset=['value'])

    scgenome.cnplot.plot_cell_cn_profile(
        ax, cn_data, 'value', cn_field_name=cn_field_name, max_cn=max_cn,
        chromosome=chromosome, s=s, squashy=squashy, rawy=rawy)

    return ax



import logging
import matplotlib
import seaborn
import random
import scipy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as dst
from sklearn.decomposition import PCA
from matplotlib.colors import ListedColormap

from scgenome import refgenome
from scgenome import utils
from scgenome import cncluster


def hex_to_rgb(h):
    if h is None:
        return np.array((0, 0, 0))
    h = h.lstrip('#')
    return np.array(tuple(np.uint8(int(h[i:i+2], 16)) for i in (0, 2 ,4)))


color_reference = {0:'#3182BD', 1:'#9ECAE1', 2:'#CCCCCC', 3:'#FDCC8A', 4:'#FC8D59', 5:'#E34A33', 6:'#B30000', 7:'#980043', 8:'#DD1C77', 9:'#DF65B0', 10:'#C994C7', 11:'#D4B9DA'}


def get_cn_cmap(cn_data):
    min_cn = int(cn_data.min())
    max_cn = int(cn_data.max())
    assert min_cn - cn_data.min() == 0
    assert max_cn - cn_data.max() == 0
    color_list = []
    for cn in range(min_cn, max_cn+1):
        if cn > max(color_reference.keys()):
            cn = max(color_reference.keys())
        color_list.append(color_reference[cn])
    return ListedColormap(color_list)


def plot_cbar(ax):
    ax.imshow(np.array([np.arange(len(color_reference))]).T[::-1], cmap=cmap, aspect=1)
    ax.set_xticks([])
    ax.set_yticks(np.arange(len(color_reference)))
    ax.set_yticklabels(np.arange(len(color_reference))[::-1])


def _secondary_clustering(data):
    D = dst.squareform(dst.pdist(data.T, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1, no_plot=True)
    idx = np.array(Z['leaves'])
    ordering = np.zeros(idx.shape[0], dtype=int)
    ordering[idx] = np.arange(idx.shape[0])
    return ordering


def plot_clustered_cell_cn_matrix(ax, cn_data, cn_field_name, cluster_field_name='cluster_id', secondary_field_name=None, raw=False, max_cn=13, cmap=None):
    plot_data = cn_data.merge(refgenome.info.chromosome_info[['chr', 'chr_index']])

    if secondary_field_name is not None:
        plot_data = plot_data.set_index(['chr_index', 'start', 'cell_id', cluster_field_name])
        plot_data = plot_data[[secondary_field_name, cn_field_name]]
        plot_data = plot_data.unstack(level=['cell_id', cluster_field_name]).fillna(0)
        ordering = plot_data[secondary_field_name].values[0]

        plot_data = cn_data.merge(refgenome.info.chromosome_info[['chr', 'chr_index']])
        plot_data = plot_data.set_index(['chr_index', 'start', 'cell_id', cluster_field_name])[cn_field_name].unstack(level=['cell_id', cluster_field_name]).fillna(0)
    else:
        plot_data = plot_data.set_index(['chr_index', 'start', 'cell_id', cluster_field_name])[cn_field_name].unstack(level=['cell_id', cluster_field_name]).fillna(0)
        ordering = _secondary_clustering(plot_data.values)
    ordering = pd.Series(ordering, index=plot_data.columns, name='cell_order')
    plot_data = plot_data.T.set_index(ordering, append=True).T

    plot_data = plot_data.sort_index(axis=1, level=[1, 2])
    if max_cn is not None:
        plot_data[plot_data > max_cn] = max_cn

    mat_chrom_idxs = plot_data.index.get_level_values(0).values
    chrom_boundaries = np.array([0] + list(np.where(mat_chrom_idxs[1:] != mat_chrom_idxs[:-1])[0]) + [plot_data.shape[0] - 1])
    chrom_sizes = chrom_boundaries[1:] - chrom_boundaries[:-1]
    chrom_mids = chrom_boundaries[:-1] + chrom_sizes / 2
    ordered_mat_chrom_idxs = mat_chrom_idxs[np.where(np.array([1] + list(np.diff(mat_chrom_idxs))) != 0)]
    chrom_names = np.array(refgenome.info.plot_chromosomes)[ordered_mat_chrom_idxs]

    mat_cluster_ids = plot_data.columns.get_level_values(1).values
    cluster_boundaries = np.array([0] + list(np.where(mat_cluster_ids[1:] != mat_cluster_ids[:-1])[0]) + [plot_data.shape[1] - 1])
    cluster_sizes = cluster_boundaries[1:] - cluster_boundaries[:-1]
    cluster_mids = cluster_boundaries[:-1] + cluster_sizes / 2

    if not raw and cmap is None:
        cmap = get_cn_cmap(plot_data.values)

    im = ax.imshow(plot_data.astype(float).T, aspect='auto', cmap=cmap, interpolation='none')

    ax.set(xticks=chrom_mids)
    ax.set(xticklabels=chrom_names)

    for val in chrom_boundaries[:-1]:
        ax.axvline(x=val, linewidth=1, color='black', zorder=100)

    return plot_data


def plot_clustered_cell_cn_matrix_figure(fig, cn_data, cn_field_name, cluster_field_name='cluster_id', secondary_field_name=None, raw=False, max_cn=13, cmap=None):
    ax = fig.add_axes([0.1,0.0,0.9,1.])
    plot_data = plot_clustered_cell_cn_matrix(ax, cn_data, cn_field_name, cluster_field_name=cluster_field_name, secondary_field_name=secondary_field_name,raw=raw, max_cn=max_cn, cmap=cmap)

    cluster_ids = plot_data.columns.get_level_values(1).values
    color_mat = cncluster.get_cluster_colors(cluster_ids)

    ax = fig.add_axes([0.0,0.0,0.05,1.])
    ax.imshow(np.array(color_mat)[::-1, np.newaxis], aspect='auto', origin='lower', interpolation='none')
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])

    return plot_data


def plot_cell_cn_profile(ax, cn_data, value_field_name, cn_field_name=None, max_cn=13, chromosome=None, s=5, squashy=False, rawy=False, cmap=None):
    """ Plot copy number profile on a genome axis

    Args:
        ax: matplotlib axis
        cn_data: copy number table
        value_field_name: column in cn_data to use for the y axis value
    
    Kwargs:
        cn_field_name: state column to color scatter points
        max_cn: max copy number for y axis
        chromosome: single chromosome plot
        s: size of scatter points
        squashy: compress y axis
        rawy: raw data on y axis

    The cn_data table should have the following columns (in addition to value_field_name and
    optionally cn_field_name):
        - chr
        - start
        - end
    """
    chromosome_info = refgenome.info.chromosome_info[['chr', 'chromosome_start', 'chromosome_end']].copy()
    chromosome_info['chr'] = pd.Categorical(chromosome_info['chr'], categories=cn_data['chr'].cat.categories)
    plot_data = cn_data.merge(chromosome_info)
    plot_data = plot_data[plot_data['chr'].isin(refgenome.info.chromosomes)]
    plot_data['start'] = plot_data['start'] + plot_data['chromosome_start']
    plot_data['end'] = plot_data['end'] + plot_data['chromosome_start']

    squash_coeff = 0.15
    squash_f = lambda a: np.tanh(squash_coeff * a)
    if squashy:
        plot_data[value_field_name] = squash_f(plot_data[value_field_name])

    if cn_field_name is not None:
        if cmap is not None:
            ax.scatter(
                plot_data['start'], plot_data[value_field_name],
                c=plot_data[cn_field_name], s=s,
                cmap=cmap,
            )
        else:
            ax.scatter(
                plot_data['start'], plot_data[value_field_name],
                c=plot_data[cn_field_name], s=s,
                cmap=get_cn_cmap(plot_data[cn_field_name].astype(int).values),
            )
    else:
        ax.scatter(
            plot_data['start'], plot_data[value_field_name], s=s,
        )

    if chromosome is not None:
        chromosome_length = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_length']
        chromosome_start = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_start']
        chromosome_end = refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_end']
        xticks = np.arange(0, chromosome_length, 2e7)
        xticklabels = ['{0:d}M'.format(int(x / 1e6)) for x in xticks]
        xminorticks = np.arange(0, chromosome_length, 1e6)
        ax.set_xlabel(f'chromosome {chromosome}')
        ax.set_xticks(xticks + chromosome_start)
        ax.set_xticklabels(xticklabels)
        ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(xminorticks + chromosome_start))
        ax.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.set_xlim((chromosome_start, chromosome_end))

    else:
        ax.set_xlim((-0.5, refgenome.info.chromosome_info['chromosome_end'].max()))
        ax.set_xlabel('chromosome')
        ax.set_xticks([0] + list(refgenome.info.chromosome_info['chromosome_end'].values))
        ax.set_xticklabels([])
        ax.xaxis.tick_bottom()
        ax.yaxis.tick_left()
        ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(refgenome.info.chromosome_info['chromosome_mid']))
        ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(refgenome.info.plot_chromosomes))

    if squashy and not rawy:
        yticks = np.array([0, 2, 4, 7, 20])
        yticks_squashed = squash_f(yticks)
        ytick_labels = [str(a) for a in yticks]
        ax.set_yticks(yticks_squashed)
        ax.set_yticklabels(ytick_labels)
        ax.set_ylim((-0.01, 1.01))
        ax.spines['left'].set_bounds(0, 1)
    elif not rawy:
        ax.set_ylim((-0.05*max_cn, max_cn))
        ax.set_yticks(range(0, int(max_cn) + 1))
        ax.spines['left'].set_bounds(0, max_cn)

    if chromosome is not None:
        seaborn.despine(ax=ax, offset=10, trim=False)
    else:
        seaborn.despine(ax=ax, offset=10, trim=True)

    return chromosome_info


def plot_breakends(ax, breakends, lw=0.5):
    """ Plot breakpoint flags and arcs on a genome axis

    Args:
        ax: matplotlib axis
        breakends: breakend table
    
    Kwargs:
        lw: line width for arcs

    The breakends table should have the following columns:
        - prediction_id
        - chromosome
        - position
        - strand

    """
    chromosome_info = refgenome.info.chromosome_info[['chr', 'chromosome_start']].copy()
    chromosome_info['chromosome'] = chromosome_info['chr']

    plot_data = breakends.merge(chromosome_info)
    plot_data = plot_data[plot_data['chr'].isin(refgenome.info.chromosomes)]
    plot_data['plot_position'] = plot_data['position'] + plot_data['chromosome_start']

    xlim = ax.get_xlim()
    xrng = xlim[1] - xlim[0]

    ylim = ax.get_ylim()
    yrng = ylim[1] - ylim[0]

    arrow_length = 0.01 * xrng

    head_width = yrng / 30.
    head_length = xrng / 100.

    prediction_ids = list(plot_data['prediction_id'].unique())
    random.shuffle(prediction_ids)
    color_palette = seaborn.color_palette('hls', len(prediction_ids))
    prediction_colors = dict(zip(prediction_ids, color_palette))

    for idx in plot_data.index:
        prediction_id = plot_data.loc[idx, 'prediction_id']
        x = plot_data.loc[idx, 'plot_position']
        strand = plot_data.loc[idx, 'strand']
        y = np.random.uniform(ylim[0] + 0.75*yrng, ylim[0] + 0.95*yrng)
        offset = (arrow_length, -arrow_length)[strand == '+']
        color = prediction_colors[prediction_id]
        ax.arrow(x, y, offset, 0, color=color, lw=lw, alpha=1.0, head_width=head_width, head_length=head_length)
        ax.plot([x, x], [ylim[0], ylim[1]], color=color, lw=lw, ls='-', alpha=1.0)

    for prediction_id, pred_breakends in plot_data.groupby('prediction_id'):
        if len(pred_breakends.index) != 2:
            continue
        pos1 = pred_breakends['plot_position'].min()
        pos2 = pred_breakends['plot_position'].max()
        posmid = (pos1 + pos2) / 2.
        height = 0.5 * yrng * (pos2 - pos1) / xrng
        ypos = ylim[1]
        yposmid = ypos + height
        spl = scipy.interpolate.make_interp_spline([pos1, posmid, pos2], [ypos, yposmid, ypos], k=2)
        pos = np.linspace(pos1, pos2, 100)
        color = prediction_colors[prediction_id]
        ax.plot(pos, spl(pos), ls='-', lw=lw, color=color)

    ax.set_ylim((ylim[0], ylim[1] + 5))


def plot_cluster_cn_matrix(fig, cn_data, cn_field_name, cluster_field_name='cluster_id'):
    plot_data = cn_data.merge(refgenome.info.chromosome_info[['chr', 'chr_index']])
    plot_data = plot_data.groupby(['chr_index', 'start', cluster_field_name], observed=True)[cn_field_name].median().astype(int)
    plot_data = plot_data.unstack(level=[cluster_field_name]).fillna(0)
    plot_data = plot_data.sort_index(axis=1, level=1)

    logging.info(f'matrix with size {plot_data.T.shape}')

    ax = fig.add_axes([0.0,1.,0.1,1.])
    if plot_data.shape[1] > 2:
        D = dst.squareform(dst.pdist(plot_data.T, 'cityblock'))
        Y = sch.linkage(D, method='complete')
        Z = sch.dendrogram(Y, color_threshold=-1, orientation='left')
        idx = Z['leaves'][::-1]
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        plot_data = plot_data.iloc[:, idx]

    mat_chrom_idxs = plot_data.index.get_level_values(0).values
    chrom_boundaries = np.array([0] + list(np.where(mat_chrom_idxs[1:] != mat_chrom_idxs[:-1])[0]) + [plot_data.shape[0] - 1])
    chrom_sizes = chrom_boundaries[1:] - chrom_boundaries[:-1]
    chrom_mids = chrom_boundaries[:-1] + chrom_sizes / 2
    chromosomes = refgenome.info.chromosome_info.set_index('chr_index').loc[np.unique(mat_chrom_idxs), 'chr_plot']

    ax = fig.add_axes([0.125,1.,0.875,1.])
    im = ax.imshow(plot_data.T, aspect='auto', cmap=get_cn_cmap(plot_data.values), interpolation='none')

    ax.set(xticks=chrom_mids)
    ax.set(xticklabels=chromosomes)
    ax.set(yticks=list(range(len(plot_data.columns.values))))
    ax.set(yticklabels=plot_data.columns.values)

    for val in chrom_boundaries[:-1]:
        ax.axvline(x=val, linewidth=1, color='black', zorder=100)

    return plot_data.columns.values


def compute_pca_loadings(cn_data):
    """ Compute the first n components of a PCA
    """
    cn_matrix = (
        cn_data
        .set_index(['cell_id', 'chr', 'start', 'end'])['copy']
        .unstack(level=[1, 2, 3]))

    num_null = cn_matrix.isnull().sum(axis=1)
    cn_matrix = cn_matrix[num_null <= 800]
    cn_matrix = cn_matrix.dropna(axis='columns')

    pca = PCA()
    pca.fit(cn_matrix.values)

    components = pd.DataFrame(
        pca.components_,
        columns=cn_matrix.columns)

    return components


def plot_pca_components(cn_data, n_components=4, plots_prefix=None):
    """ Plot the first n components of a PCA
    """
    components = compute_pca_loadings(cn_data)

    fig = plt.figure(figsize=(20, 4 * n_components))
    for idx in range(4):
        ax = fig.add_subplot(n_components, 1, idx + 1)
        plot_data = components.iloc[idx].T.rename('component').reset_index()
        plot_cell_cn_profile(
            ax, plot_data, 'component', rawy=True)
        ax.set_ylabel(f'PCA {idx+1}')

    if plots_prefix is not None:
        fig.savefig(plots_prefix + 'pca_components.pdf', bbox_inches='tight')




import logging
import numpy as np
import scipy.stats

from anndata import AnnData



_default_filters = (
    'filter_quality',
    'filter_reads',
    'filter_copy_state_diff',
    'filter_is_s_phase',
)

def calculate_filter_metrics(
        adata: AnnData,
        quality_score_threshold=0.75,
        read_count_threshold=500000,
        copy_state_diff_threshold=1.,
        inplace = False,
    ) -> AnnData:
    """ Calculate additional filtering metrics to be used by other filtering methods.

    Parameters
    ----------
    adata : AnnData
        copy number data on which to calculate filter metrics
    quality_score_threshold : float, optional
        The minimum quality to set to keep, by default 0.75
    read_count_threshold : int, optional
        The minimum total mapped reads from hmmcopy to set for keeping, by default 500000
    copy_state_diff_threshold : [type], optional
        Minimum copy-state difference threshold to set to keep, by default 1.
    inplace : bool, optional
        Whether to modify passed in AnnData, by default False

    Returns
    -------
    AnnData
        AnnData with modified obs if not inplace, otherise, None

    Note
    ----

    The following properties are changed:
    - AnnData.obs.filter_quality
    - AnnData.obs.filter_reads
    - AnnData.obs.filter_copy_state_diff
    
    If is_s_phase is a property of AnnData
        AnnData.obs.filter_is_s_phase
        
    AnnData.obs.copy_state_diff
    AnnData.obs.copy_state_diff_mean
    """
    if not inplace:
        ndad = adata.copy()
        
        return calculate_filter_metrics(
            ndad,
            quality_score_threshold,
            read_count_threshold,
            copy_state_diff_threshold,
            inplace = True,
        )

    # Filter Quality and Filter Reads
    if 'quality' in adata.obs.columns:
        adata.obs['filter_quality'] = (adata.obs['quality'] > quality_score_threshold)
    else:
        logging.warning("quality is not in AnnData.obs. Skipping filter_quality")
    
    if 'total_mapped_reads_hmmcopy' in adata.obs.columns:
        adata.obs['filter_reads'] = (adata.obs['total_mapped_reads_hmmcopy'] > read_count_threshold)
    else:
        logging.warning("total_mapped_reads_hmmcopy is not in AnnData.obs. Skipping total_mapped_reads_hmmcopy")
    
    # Copy State Difference Filter
    adata.obsm['copy_state_diff'] = np.absolute(adata.layers['copy'] - adata.layers['state'])
    adata.obsm['copy_state_diff_mean'] = np.nanmean(adata.obsm['copy_state_diff'], axis=1)

    adata.obs['filter_copy_state_diff'] = (adata.obsm['copy_state_diff_mean'] < copy_state_diff_threshold)

    # Filter s phase column
    if 'is_s_phase' in adata.obs.columns:
        adata.obs['filter_is_s_phase'] = ~(adata.obs['is_s_phase'].fillna(False))
    else:
        logging.warning("No is_s_phase in AnnData.obs. Skipping filter_is_s_phase")

    return adata


def filter_cells(
        adata: AnnData,
        filters = _default_filters,
        inplace = False,
    ) -> AnnData:
    """
    Filter poor quality cells based on the filters provided.

    Parameters
    -------
    adata : AnnData
        AnnData to preform operation with
    filters : list, optional
        Filters to apply. Keeps cells where filters are true, by default _default_filters
    inplace
        Whether to modify passed in AnnData. If False, returns new AnnData.

    Returns
    -------
    AnnData
        filtered copy number data
    """

    if not inplace:
        adata = adata.copy()
        
        return filter_cells(
            adata,
            filters,
            inplace = True,
        )

    # Ensure cnfilter.calculate_filter_metrics has been called
    for filter_option in filters:
        if filter_option not in adata.obs.columns:
            logging.warning(
                f"WARNING: {filter_option} is not found! "
                "Skipping. Are you sure `scgenome.pp.calculate_filter_metrics` has been called?"
            )
            continue

        adata = adata[adata.obs[filter_option]]

    return adata





import os
from collections import defaultdict

import pandas as pd
import scgenome.loaders.utils
import scgenome.utils
import yaml
from csverve.core import CsverveInput


standard_hmmcopy_reads_cols = [
    'chr',
    'start',
    'end',
    'cell_id',
    'gc',
    'reads',
    'copy',
    'state',
]

_categorical_cols = [
    'cell_id',
    'chr',
    'sample_id',
    'library_id',
]


def load_hmmcopy_files(
        hmmcopy_reads, hmmcopy_segs, hmmcopy_metrics,
        additional_reads_cols=None
    ):
    results_tables = {}

    hmmcopy_reads_cols = standard_hmmcopy_reads_cols.copy()
    if additional_reads_cols is not None:
        hmmcopy_reads_cols.extend(additional_reads_cols)

    results_tables["hmmcopy_reads"] = process_hmmcopy_data(hmmcopy_reads, usecols=hmmcopy_reads_cols)
    results_tables["hmmcopy_segs"] = process_hmmcopy_data(hmmcopy_segs)
    results_tables["hmmcopy_metrics"] = process_hmmcopy_data(hmmcopy_metrics)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def load_hmmcopy_results(
        results_dir,
        additional_reads_cols=None,
    ):
    """ Load copy number tables
    
    Args:
        results_dir (str): results directory to load from.

    KwArgs:
        additional_reads_cols (list of str, optional): Additional columns to obtain from the reads table. Defaults to None.
    
    Returns:
        dict: pandas.DataFrame tables keyed by table name
    """

    hmmcopy_reads_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_reads.csv.gz', 'hmmcopy_reads', analysis_type='hmmcopy')

    hmmcopy_segs_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_segments.csv.gz', 'hmmcopy_segments', analysis_type='hmmcopy')

    hmmcopy_metrics_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, '_metrics.csv.gz', 'hmmcopy_metrics', analysis_type='hmmcopy')

    return load_hmmcopy_files(
        hmmcopy_reads_filepath, hmmcopy_segs_filepath, hmmcopy_metrics_filepath,
        additional_reads_cols=additional_reads_cols,
    )


def process_hmmcopy_data(filepath, usecols=None):
    data = CsverveInput(filepath).read_csv(usecols=usecols)

    data['sample_id'] = [a.split('-')[-4] for a in data['cell_id']]
    data['library_id'] = [a.split('-')[-3] for a in data['cell_id']]

    for col in _categorical_cols:
        if col in data:
            data[col] = pd.Categorical(data[col])

    return data



import logging
import seaborn

import numpy as np
import pandas as pd
import scipy.stats
import scipy.misc
import scipy.special
import matplotlib.pyplot as plt
import matplotlib

import dollo.tasks
import dollo.run
import wgs_analysis.plots.trees

import scgenome.snvphylo


def annotate_copy_number(pos, seg, columns=['major', 'minor'], sample_col='sample_id'):
    """ Annotate positions with segment specific data 
    """
    results = []

    for sample_id in seg[sample_col].unique():
        sample_pos = pos[pos[sample_col] == sample_id]
        sample_seg = seg[seg[sample_col] == sample_id]

        for chrom in seg['chr'].unique():
            _pos = sample_pos[sample_pos['chrom'] == chrom]
            _seg = sample_seg[sample_seg['chr'] == chrom]

            results.append(find_overlapping_segments(_pos, _seg, columns))

    return pd.concat(results)


def find_overlapping_segments(pos, seg, columns):
    """ Find positions that are contained within segments
    """
    seg = seg.sort_values(['start', 'end'])

    if seg.duplicated(['start', 'end']).any():
        raise ValueError('duplicate columns')

    start_idx = np.searchsorted(seg['start'].values, pos['coord'].values) - 1
    end_idx = np.searchsorted(seg['end'].values, pos['coord'].values)

    mask = (start_idx == end_idx)

    results = pos.copy()

    for col in columns:
        results[col] = np.nan
        results.loc[mask, col] = seg[col].iloc[end_idx[mask]].values

    return results


def snv_hierarchical_clustering_figure(snv_data, clusters):
    """ Simple hierarhical clustering figure for SNVs
    """
    snv_matrix = snv_data.merge(clusters)

    snv_matrix = (
        snv_matrix.groupby(
            ['chrom', 'coord', 'ref', 'alt', 'cluster_id'],
            as_index=True, observed=True)[['alt_counts', 'ref_counts']]
        .sum().unstack().fillna(0).astype(int).stack().reset_index())
    snv_matrix['total_counts'] = snv_matrix['ref_counts'] + snv_matrix['alt_counts']

    snv_matrix['vaf'] = snv_matrix['alt_counts'] / snv_matrix['total_counts']
    snv_matrix['alt_counts'] = snv_matrix['alt_counts'].clip(upper=10)
    snv_matrix['is_present'] = (snv_matrix['alt_counts'] > 0) * 1
    snv_matrix['is_absent'] = (snv_matrix['alt_counts'] == 0) * 1
    snv_matrix['is_het'] = (snv_matrix['alt_counts'] < 0.99 * snv_matrix['total_counts']) * snv_matrix['is_present']
    snv_matrix['is_hom'] = (snv_matrix['alt_counts'] >= 0.99 * snv_matrix['total_counts']) * snv_matrix['is_present']
    snv_matrix['state'] = snv_matrix['is_hom'] * 3 + snv_matrix['is_het'] * 2 + snv_matrix['is_absent']
    snv_presence_matrix = snv_matrix.set_index(['chrom', 'coord', 'cluster_id'])['is_present'].unstack(fill_value=0)

    logging.info(f'snv matrix with shape {snv_presence_matrix.shape}, memory {snv_presence_matrix.memory_usage().sum()}')

    # KLUDGE: currently recursion in dendrograms
    # breaks with large datasets
    import sys
    sys.setrecursionlimit(10000)

    g = seaborn.clustermap(snv_presence_matrix, rasterized=True, row_cluster=True, figsize=(4, 12))

    return g.fig


def compute_snv_log_likelihoods(snv_data, allele_cn, clusters):
    """ Compute log likelihoods of presence absence for SNVs
    """
    snv_matrix = snv_data.merge(clusters)

    snv_matrix = (
        snv_matrix.groupby(
            ['chrom', 'coord', 'ref', 'alt', 'cluster_id'],
            as_index=True, observed=True)[['alt_counts', 'ref_counts']]
        .sum().unstack().fillna(0).astype(int).stack().reset_index())
    snv_matrix['total_counts'] = snv_matrix['ref_counts'] + snv_matrix['alt_counts']

    snv_matrix['variant_id'] = snv_matrix.apply(
        lambda row: ':'.join(row[['chrom', 'coord', 'ref', 'alt']].astype(str).values),
        axis=1).astype('category')

    # TODO: this should be moved
    allele_cn['total_cn'] = allele_cn['total_cn'].astype(int)
    allele_cn['minor_cn'] = allele_cn['minor_cn'].astype(int)
    allele_cn['major_cn'] = allele_cn['major_cn'].astype(int)
    allele_cn = allele_cn[[
        'chr', 'start', 'end', 'cluster_id',
        'total_cn', 'minor_cn', 'major_cn',
    ]].drop_duplicates()

    # Merge segment copy number into SNV table
    snv_log_likelihoods = annotate_copy_number(
        snv_matrix, allele_cn,
        columns=['major_cn', 'minor_cn', 'total_cn'],
        sample_col='cluster_id')

    snv_log_likelihoods = compute_log_likelihoods(
        snv_log_likelihoods)

    return snv_log_likelihoods


def compute_log_likelihoods(df, error_rate=1e-3):
    """ Compute the presence absence log likelihood of an SNV
    """
    df['log_likelihood_absent'] = df.apply(calculate_likelihood_absent, axis=1, args=(error_rate,))
    df['log_likelihood_present'] = df.apply(calculate_likelihood_present, axis=1, args=(error_rate,))

    return df


def calculate_likelihood_absent(row, e_s):
    return log_likelihood_absent(
        e_s,
        row['alt_counts'],
        row['alt_counts'] + row['ref_counts'],
    )


def calculate_likelihood_present(row, e_s):
    return log_likelihood_present(
        row['major_cn'],
        row['major_cn'] + row['minor_cn'],
        e_s,
        row['alt_counts'],
        row['ref_counts'] + row['alt_counts'],
    )

def log_binomial_pdf(x, n, p):
    return scipy.stats.binom.logpmf(x, n, p)


def log_likelihood_absent(e_s, n_v, n_t):
    return log_binomial_pdf(n_v, n_t, e_s)


def log_likelihood_present(c_m, c_t, e_s, n_v, n_t):
    if c_m == 0:
        return log_likelihood_absent(e_s, n_v, n_t)

    conditional_log_likelihoods = []

    for c_v in np.arange(1., c_m + 1., 1.):
        allele_ratio = c_v / c_t
        r = (1 - e_s) * allele_ratio + e_s * (1 - allele_ratio)
        conditional_log_likelihoods.append(log_binomial_pdf(n_v, n_t, r))

    return scipy.special.logsumexp(conditional_log_likelihoods)


def compute_dollo_ml_tree(snv_log_likelihoods, leaf_name_groups=None):
    """ Compute the ML tree under the dollo model of SNV evolution
    """
    trees = dollo.tasks.create_trees(
        snv_log_likelihoods,
        sample_col='cluster_id',
        leaf_name_groups=leaf_name_groups,
    )

    results_table = dollo.tasks.compute_tree_log_likelihoods_mp(
        snv_log_likelihoods, trees,
        sample_col='cluster_id', variant_col='variant_id')

    ml_tree_id = results_table.set_index('tree_id')['log_likelihood'].idxmax()
    tree = trees[ml_tree_id]
    loss_prob = results_table.set_index('tree_id').loc[ml_tree_id, 'loss_prob']

    tree_annotations = dollo.run.annotate_posteriors(
        snv_log_likelihoods, tree, loss_prob=loss_prob,
        sample_col='cluster_id', variant_col='variant_id')

    return tree, tree_annotations


def plot_dollo_ml_tree(tree, nodes):
    """ Plot a tree with branchlengths as snv origin counts
    """
    leaf_order = []
    for leaf in tree.leaves:
        leaf.plot_id = leaf.name
        leaf_order.append(leaf.name)

    origin_counts = nodes.groupby('node')['ml_origin'].sum()

    for node in tree.nodes:
        node.origin_count = origin_counts[node.label]

    loss_counts = nodes.groupby('node')['ml_loss'].sum()

    width = 1 + 0.5 * float(len(list(tree.leaves)))
    fig = plt.figure(figsize=(width/1.5, 6))

    ax = fig.add_subplot(111)

    def func(x, pos):
        s = '{:0,d}'.format(int(x))
        return s
    ax.yaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(func))

    wgs_analysis.plots.trees.plot_tree(ax, tree, landscape=False, flip=True, branch_length_attr='origin_count', leaf_name_attr='plot_id')

    ax.set_ylabel('SNV count')

    plt.tight_layout()


def run_snv_phylogenetics(snv_count_data, allele_cn, clusters, results_prefix):
    """ Run the SNV phylogenetic analysis.
    """
    snv_log_likelihoods = scgenome.snvphylo.compute_snv_log_likelihoods(
        snv_count_data, allele_cn, clusters)

    ml_tree, tree_annotations = scgenome.snvphylo.compute_dollo_ml_tree(
        snv_log_likelihoods)

    return ml_tree, tree_annotations




import os

import pandas as pd
import scgenome.loaders.utils
import yaml
import scgenome.loaders.align
import scgenome.loaders.annotation
import scgenome.loaders.hmmcopy


def load_qc_files(
        hmmcopy_reads,
        hmmcopy_segs,
        hmmcopy_metrics,
        alignment_metrics,
        gc_metrics,
        annotation_metrics,
        sample_ids=None,
        additional_hmmcopy_reads_cols=None,
    ):

    results_tables = scgenome.loaders.align.load_alignment_files(
        alignment_metrics,
        gc_metrics=gc_metrics
    )

    hmmcopy_results_tables = scgenome.loaders.hmmcopy.load_hmmcopy_files(
        hmmcopy_reads,
        hmmcopy_segs, hmmcopy_metrics,
        additional_reads_cols=additional_hmmcopy_reads_cols
    )
    results_tables.update(hmmcopy_results_tables)

    annotation_results_tables = scgenome.loaders.annotation.load_annotation_files(annotation_metrics)
    results_tables.update(annotation_results_tables)

    if sample_ids is not None:
        results_tables = _sample_id_filter(results_tables, sample_ids)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def load_qc_results(
        alignment_results_dir,
        hmmcopy_results_dir,
        annotation_results_dir=None,
        sample_ids=None,
        additional_hmmcopy_reads_cols=None,
):
    """ Load qc data (align, hmmcopy)
    
    Args:
        alignment_results_dir (str): alignment results directory to load from.
        hmmcopy_results_dir (str): hmmcopy results directory to load from.

    KwArgs:
        annotation_results_dir (str): annotation results directory to load from. Defaults to None.
        sample_ids (list of str, optional): Set of sample ids to filter for. Defaults to None.
        additional_hmmcopy_reads_cols (list of str, optional): Additional columns to obtain from the reads table. Defaults to None.
    """

    alignment_results_dir = scgenome.loaders.utils.find_results_directory(
        alignment_results_dir, 'alignment')

    results_tables = scgenome.loaders.align.load_alignment_results(alignment_results_dir)

    hmmcopy_results_dir = scgenome.loaders.utils.find_results_directory(
        hmmcopy_results_dir, 'hmmcopy')

    hmmcopy_results_tables = scgenome.loaders.hmmcopy.load_hmmcopy_results(
        hmmcopy_results_dir,
        additional_reads_cols=additional_hmmcopy_reads_cols)

    results_tables.update(hmmcopy_results_tables)

    if annotation_results_dir is not None:
        annotation_results_dir = scgenome.loaders.utils.find_results_directory(
            annotation_results_dir, 'annotation')

        annotation_results_tables = scgenome.loaders.annotation.load_annotation_results(annotation_results_dir)

        results_tables.update(annotation_results_tables)

    if sample_ids is not None:
        results_tables = _sample_id_filter(results_tables, sample_ids)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def _sample_id_filter(results_tables, sample_ids):
    for table_name, table_data in results_tables.items():
        results_tables[table_name] = table_data[table_data['sample_id'].isin(sample_ids)]

    return results_tables



import logging

import pandas as pd
import scgenome.csvutils
import scgenome.loaders.utils
import scgenome.utils


def load_haplotype_allele_counts_results(results_dir):
    """ Load haplotype allele counts from a results directory.

    Args:
        results_dir (str): results directory
    """

    allele_counts_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'allele_counts.csv.gz', analysis_type='count_haps')

    return load_haplotype_allele_counts_files(allele_counts_filepath)


def load_haplotype_allele_counts_files(allele_counts_filepath):
    """ Load haplotype allele counts from a file.

    Args:
        allele_counts_filepath (str): results filepath
    """

    csv_input = scgenome.csvutils.CsvInput(allele_counts_filepath)
    data = csv_input.read_csv(
        dtypes_override={
            'chromosome': 'category',
            'cell_id': 'category',
    })

    return data




import Bio.Phylo
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import gridspec
import seaborn as sns

import scgenome.plotting.heatmap


def map_annotations_to_colors(annotation, cmap):
    num_colors = len(np.unique(annotation))

    color_map = {}
    for idx, value in enumerate(np.unique(annotation)):
        color_map[value] = cmap(float(idx) / float(max(1, num_colors - 1)))

    color_mat = []
    for value in annotation:
        color_mat.append(color_map[value])
        
    return color_mat, color_map


def plot_tree_cn(
        tree,
        adata,
        chrom_segments=True,
        layer_name=None,
        obs_annotation=None,
        obs_cmap=None,
        var_label=None,
        fig=None,
        raw=False,
        max_cn=13):
    """ Plot a tree aligned to a CN values matrix heatmap

    Parameters
    ----------
    tree : Bio.Phylo.BaseTree.Tree
        phylogenetic tree
    adata : AnnData
        Copy number data, either genes or segments
    chrom_segments : bool, optional
        whether adata is genes or segments, by default True
    layer_name : str, optional
        layer to plot for copy number heatmap, by default None
    obs_annotation : str, optional
        column of adata.obs to annotate cells, by default None
    obs_cmap : matplotlib.colors.ListedColormap, optional
        color map for cell annotations, by default None
    var_label : str, optional
        column of adata.var to use for heatmap var labels, by default None use index
    fig : matplotlib.figure.Figure, optional
        existing figure to plot into, by default None
    raw : bool, optional
        raw plotting, no integer color map, by default False
    max_cn : int, optional
        clip cn at max value, by default 13
    """    
    
    if fig is None:
        fig = plt.figure(figsize=(16, 12), dpi=150)

    # Add phylogenetic ordering to anndata obs
    cell_ids = []
    for a in tree.get_terminals():
        cell_ids.append(a.name)

    adata.obs['phylo_order'] = None
    for idx, _ in adata.obs.iterrows():
        adata.obs.loc[idx, 'phylo_order'] = cell_ids.index(idx)

    gs = gridspec.GridSpec(1, 3, width_ratios=(0.4, 0.58, 0.02))

    # Plot phylogenetic tree
    ax = fig.add_subplot(gs[0, 0])
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(False)
    ax.get_yaxis().set_ticks([])
    Bio.Phylo.draw(tree, label_func=lambda a: '', axes=ax, do_show=False)

    # Plot copy number heatmap
    ax = fig.add_subplot(gs[0, 1])

    if chrom_segments:
        scgenome.plotting.heatmap.plot_cell_cn_matrix(
            adata,
            layer_name=layer_name,
            cell_order_fields=('phylo_order',),
            ax=ax,
            raw=raw,
            max_cn=max_cn,
        )

    else:
        if layer_name is not None:
            X = adata.layers[layer_name]
        else:
            X = adata.X

        # Order the cells according to the phylogeny
        cell_order_values = adata.obs[['phylo_order']].values.transpose()
        cell_ordering = np.lexsort(cell_order_values)

        X = X[cell_ordering, :]

        mat = adata[cell_ordering, :].to_df(layer=layer_name)
        if var_label is not None:
            mat = mat.rename(columns=adata.var[var_label])
        
        cbar_ax = fig.add_axes([0.45, 0.0, 0.2, 0.01])
        sns.heatmap(mat, ax=ax, cbar_ax=cbar_ax, cbar_kws={'orientation': 'horizontal'})

    ax.set_yticks([])

    # Plot obs annotation
    if obs_annotation is not None:
        ax = fig.add_subplot(gs[0, 2])

        color_mat, color_map = map_annotations_to_colors(adata.obs[obs_annotation], obs_cmap)
        color_mat = np.swapaxes(np.array([color_mat]), 0, 1)

        ax.imshow(color_mat[::-1, :, :], aspect='auto', origin='lower', interpolation='none')
        ax.set_xticks([])
        ax.set_yticks([])

        patches = []
        for label, color in color_map.items():
            patches.append(mpatches.Patch(color=color, label=label))
        ax.legend(handles=patches, loc='upper left', bbox_to_anchor=(1.05, 1.))

    plt.subplots_adjust(left=0.065, right=0.97, top=0.96, bottom=0.065, wspace=0.01)

    return fig




from .filtering import filter_cells, calculate_filter_metrics
from .load_cn import create_cn_anndata, read_dlp_hmmcopy, convert_dlp_hmmcopy, convert_dlp_signals, read_bam_bin_counts
from .load_snv import read_snv_genotyping



import pyranges as pr
import pandas as pd
import anndata as ad
import numpy as np

from pandas import DataFrame
from anndata import AnnData
from pyranges import PyRanges
from collections.abc import Iterable


def read_ensemble_genes_gtf(gtf_filename) -> PyRanges:
    """ Read an ensembl gtf and extract gene start end

    Parameters
    ----------
    gtf_filename : str
        GTF filename

    Returns
    -------
    PyRanges
        Genes bounds
    """    
    genes = pr.read_gtf(gtf_filename, as_df=True)
    genes = genes.groupby(['Chromosome', 'gene_id', 'gene_name'], observed=True).agg({'Start': min, 'End': max}).reset_index()
    genes = pr.PyRanges(genes)
    return genes


def _segment_width_weighted_mean_matrix(data: DataFrame, intersect: DataFrame) -> DataFrame:
    # Add gene id to index
    data = data.loc[:, intersect['bin']].T.set_index(pd.Index(intersect['gene_id'].values, name='gene_id'), append=True)

    # Scale by overlap segment width
    data = data * intersect['segment_width'].values[:, np.newaxis]

    # Calculate segment width normalized mean
    data = data.groupby(level=1).sum()
    data = data.T / intersect.groupby('gene_id')['segment_width'].sum()

    return data


def _segment_width_weighted_mean_var(data: DataFrame, intersect: DataFrame) -> DataFrame:
    # Add gene id to index
    data = data.loc[intersect['bin']].set_index(pd.Index(intersect['gene_id'].values, name='gene_id'), append=True)

    # Scale by overlap segment width
    data = data * intersect['segment_width'].values[:, np.newaxis]

    # Calculate segment width normalized mean
    data = data.groupby(level=1).sum()
    data = (data.T / intersect.groupby('gene_id')['segment_width'].sum()).T

    return data


def aggregate_genes(
        adata: AnnData,
        genes: PyRanges,
        agg_layers: Iterable=None,
        agg_var: Iterable=None) -> AnnData:
    """ Aggregate copy number by gene to create gene CN matrix

    Currently only does segment width weighted mean aggregation.

    Parameters
    ----------
    adata : AnnData
        copy number data
    genes : PyRanges
        gene data
    agg_layers : List, optional
        list of layers to aggregate, by default None, all layers
    agg_var : List, optional
        list of obs columns to aggregate, by default None, all columns
    cluster_col : str, optional
        column with cluster ids, by default 'cluster_id'

    Returns
    -------
    AnnData
        aggregated gene copy number
    """

    if agg_layers is None:
        agg_layers = adata.layers.keys()
    agg_layers = set(agg_layers)

    if agg_var is None:
        agg_var = set(adata.var.select_dtypes(include=np.number).columns.to_list()) - set(['chr', 'start', 'end'])
    agg_var = set(agg_var)

    bins = pr.PyRanges(adata.var.reset_index().rename(columns={
        'chr': 'Chromosome',
        'start': 'Start',
        'end': 'End',
    })[['Chromosome', 'Start', 'End', 'bin']])

    intersect_1 = genes.intersect(bins)
    intersect_2 = bins.intersect(genes)

    intersect = pd.merge(intersect_1.as_df(), intersect_2.as_df())
    intersect['segment_width'] = intersect['End'] - intersect['Start']

    X = _segment_width_weighted_mean_matrix(adata.to_df(), intersect)

    layer_data = {}
    for layer_name in agg_layers:
        layer_data[layer_name] = _segment_width_weighted_mean_matrix(adata.to_df(layer=layer_name), intersect)

    var = _segment_width_weighted_mean_var(adata.var[agg_var], intersect)

    gene_data = genes.as_df().drop_duplicates().set_index('gene_id')
    var = var.merge(gene_data, left_index=True, right_index=True, how='left')

    adata = ad.AnnData(
        X,
        obs=adata.obs,
        var=var,
        layers=layer_data,
    )

    return adata





import numpy as np
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as dst

from anndata import AnnData

import scgenome.preprocessing.transform


def sort_cells(adata: AnnData, layer_name=None) -> AnnData:
    """ Sort cells by hierarchical clustering on copy number values.

    Parameters
    ----------
    adata : AnnData
        copy number data
    layer_name : str, optional
        layer with copy number data to plot, None for X, by default None

    Returns
    -------
    AnnData
        copy number data with cell_order column added to obs
    """

    if layer_name is not None:
        X = adata.layers[layer_name]
    else:
        X = adata.X

    X = scgenome.preprocessing.transform.fill_missing(X)

    D = dst.squareform(dst.pdist(X, 'cityblock'))
    Y = sch.linkage(D, method='complete')
    Z = sch.dendrogram(Y, color_threshold=-1, no_plot=True)
    idx = np.array(Z['leaves'])

    ordering = np.zeros(idx.shape[0], dtype=int)
    ordering[idx] = np.arange(idx.shape[0])

    adata.obs['cell_order'] = ordering

    return adata





from .cluster import cluster_cells, cluster_cells_kmeans, aggregate_clusters_hmmcopy, aggregate_clusters, compute_umap
from .pca import pca_loadings
from .sorting import sort_cells
from .binfeat import create_bins, count_gc, mean_from_bigwig
from .genes import read_ensemble_genes_gtf, aggregate_genes
from .concat import ad_concat_cells
from .phylo import prune_leaves, align_cn_tree



import umap
import hdbscan
import seaborn
import logging
import itertools
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import sklearn.cluster
import scipy.spatial
from adjustText import adjust_text


def umap_hdbscan_cluster(
        cn,
        n_components=2,
        n_neighbors=15,
        min_dist=0.1,
    ):
    """ Cluster using umap and hdbscan.

    Args:
        cn: data frame columns as cell ids, rows as segments

    Returns:
        data frame with columns:
            cluster_id
            cell_id
            umap1
            umap2

    """
    embedding = umap.UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        n_components=n_components,
        random_state=42,
        metric='euclidean',
    ).fit_transform(cn.fillna(0).values.T)

    clusters = hdbscan.HDBSCAN(
        min_samples=10,
        min_cluster_size=30,
    ).fit_predict(embedding)

    df = pd.DataFrame({
        'cell_id': cn.columns, 'cluster_id': clusters,
        'umap1': embedding[:, 0], 'umap2': embedding[:, 1]
    })
    df = df[['cell_id', 'cluster_id', 'umap1', 'umap2']]
    df = df.dropna()

    return df


def compute_bic(kmeans, X):
    """ Computes the BIC metric for a given k means clustering

    Args:
        kmeans: a fitted kmeans clustering object
        X: data for which to calculate bic
    
    Returns:
        float: bic
    
    Reference: https://stats.stackexchange.com/questions/90769/using-bic-to-estimate-the-number-of-k-in-kmeans
    """
    centers = [kmeans.cluster_centers_]
    labels  = kmeans.labels_
    n_clusters = kmeans.n_clusters
    cluster_sizes = np.bincount(labels)
    N, d = X.shape

    # Compute variance for all clusters
    cl_var = (1.0 / (N - n_clusters) / d) * sum([sum(scipy.spatial.distance.cdist(X[np.where(labels == i)], [centers[0][i]], 
             'euclidean')**2) for i in range(n_clusters)])

    const_term = 0.5 * n_clusters * np.log(N) * (d+1)

    bic = np.sum([cluster_sizes[i] * np.log(cluster_sizes[i]) -
               cluster_sizes[i] * np.log(N) -
             ((cluster_sizes[i] * d) / 2) * np.log(2*np.pi*cl_var) -
             ((cluster_sizes[i] - 1) * d/ 2) for i in range(n_clusters)]) - const_term

    return bic


def kmeans_cluster(
        cn,
        min_k=2,
        max_k=100,
    ):
    """ Cluster using kmeans and bic.
    """

    X = cn.T.values
    ks = range(min_k, max_k + 1)

    logging.info(f'trying with max k={max_k}')

    kmeans = []
    bics = []
    for k in ks:
        logging.info(f'trying with k={k}')
        model = sklearn.cluster.KMeans(n_clusters=k, init="k-means++").fit(X)
        bic = compute_bic(model, X)
        kmeans.append(model)
        bics.append(bic)

    opt_k = np.array(bics).argmax()
    logging.info(f'selected k={opt_k}')

    model = kmeans[opt_k]

    embedding = umap.UMAP(
        n_neighbors=15,
        min_dist=0.1,
        n_components=2,
        random_state=42,
        metric='euclidean',
    ).fit_transform(cn.fillna(0).values.T)

    clusters = pd.DataFrame({
        'cell_id': cn.columns, 'cluster_id': model.labels_,
        'umap1': embedding[:, 0], 'umap2': embedding[:, 1]
    })

    return clusters


def get_cluster_palette(n_col):
    if n_col <= 10:
        palette = plt.get_cmap("tab10")
    elif n_col <= 21:
        palette = mpl.colors.ListedColormap([
            '#1d1d1d', '#ebce2b', '#702c8c', '#db6917', '#96cde6', '#ba1c30',
            '#c0bd7f', '#7f7e80', '#5fa641', '#d485b2', '#4277b6', '#df8461',
            '#463397', '#e1a11a', '#91218c', '#e8e948', '#7e1510', '#92ae31',
            '#6f340d', '#d32b1e', '#2b3514'
        ])
    else:
        palette = plt.get_cmap("hsv")
    return palette


def get_cluster_color_map(cluster_ids):
    num_colors = len(np.unique(cluster_ids[cluster_ids >= 0]))
    pal = get_cluster_palette(num_colors)

    color_map = {}

    cluster_ids = np.sort(np.unique(cluster_ids))

    for cluster_id in np.sort(np.unique(cluster_ids)):
        if cluster_id < 0:
            color_map[cluster_id] = (0.75, 0.75, 0.75, 1.0)

    cluster_ids = cluster_ids[cluster_ids >= 0]

    idx = 0.
    for cluster_id in itertools.chain(cluster_ids[::2], cluster_ids[1::2]):
        color_map[cluster_id] = pal(float(idx) / float(num_colors - 1))
        idx += 1

    return color_map


def get_cluster_colors(cluster_ids, color_map=None, return_map=False):
    if color_map is None:
        color_map = get_cluster_color_map(cluster_ids)

    color_mat = []
    for cluster_id in cluster_ids:
        color_mat.append(color_map[cluster_id])

    if return_map:
        return color_mat, color_map 

    return color_mat


def cluster_labels(cluster_ids):
    counts = cluster_ids.value_counts().astype(str)
    labels = counts.index.to_series().astype(str) + ' (' + counts + ')'
    if -1 in labels:
        labels.at[-1] = labels[-1].replace('-1', 'Filt.')
    return dict(list(zip(counts.index, labels)))


def plot_umap_clusters(ax, df):
    """ Scatter plot of umap clusters.

    Args:
        ax: matplotlib axis
        df: clusters dataframe with columns:
            cluster_id
            umap1
            umap2

    """
    labels = cluster_labels(df['cluster_id'])
    color_map = get_cluster_color_map(df['cluster_id'].values)

    if -1 in labels:
        df_noise = df[df['cluster_id'] < 0]
        ax.scatter(
            df_noise['umap1'].values,
            df_noise['umap2'].values,
            color=color_map[-1],
            s=2,
            label=labels[-1],
        )

    text_labels = []
    for cluster_id, cluster_df in df[df['cluster_id'] >= 0].groupby('cluster_id'):
        ax.scatter(
            cluster_df['umap1'].values,
            cluster_df['umap2'].values,
            color=color_map[cluster_id],
            s=2,
            label=labels[int(cluster_id)],
        )

    label_pos = df.groupby('cluster_id').mean()
    text_labels = [
        ax.text(label_pos.at[c, 'umap1'], label_pos.at[c, 'umap2'], c)
        for c in list(labels.keys()) if c >= 0
    ]
    adjust_text(
        text_labels, ax=ax,
        force_points=(0.1, 0.1)
    )

    ax.legend(
        frameon=False, markerscale=5,
        scatterpoints=1, bbox_to_anchor=(0.96, 0.85))
    ax.set_xlabel('Comp. 1')
    ax.set_ylabel('Comp. 2')
    seaborn.despine(ax=ax, offset=0, trim=True)





from setuptools import setup, find_packages
import versioneer

setup(
    name='scgenome',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Code for analyzing single cell whole genomes',
    author='Shah Lab',
    url='https://github.com/shahcompbio/scgenome',
    packages=find_packages(),
    install_requires=[
        'adjustText',
        'anndata',
        'bamread',
        'biopython',
        'bokeh',
        'brewer2mpl',
        'Click',
        'csverve>=0.3.1',
        'hdbscan',
        'hmmlearn',
        'ipython',
        'jupyter',
        'lda',
        'matplotlib',
        'nose',
        'numba',
        'numexpr',
        'numpy',
        'oauthlib',
        'pandas',
        'pyBigWig',
        'pyfaidx',
        'pyranges',
        'pysam',
        'PyYAML',
        'scikit-learn',
        'scipy',
        'seaborn',
        'statsmodels',
        'umap-learn',
        'wgs_analysis',
        'nose',
        'biopython',
        'pypeliner',
        'joblib==1.1.0',
        'sphinx==5.1.1'
    ],
    package_data={
        'scgenome': [
            'data/*',
            'dtypes/*.yaml',
            'datasets/data/*'
        ],
    },
)



import logging
import os

import numpy as np
import packaging
import pandas as pd
import scgenome.csvutils
import scgenome.loaders.utils
import scgenome.utils
import yaml

default_museq_filter = 0.9
default_strelka_filter = 20.
default_mappability_filter = 0.99

categorical_columns = [
    'chrom',
    'ref',
    'alt',
    'gene_name',
    'effect',
    'effect_impact',
    'amino_acid_change',
    'tri_nucleotide_context',
]


def load_snv_genotyping_results(results_dir, positions=None):
    """ Load per cell SNV count data from results directory
    
    Args:
        results_dir (str): results directory
        positions (pandas.DataFrame): restrict to the specified positions

    Returns:
        pandas.DataFrame: SNV alt and ref counts per cell
    """

    counts_filepath = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'counts.csv.gz', analysis_type='snv_genotyping')

    return load_snv_genotyping_files(counts_filepath, positions=positions)


def load_snv_genotyping_files(filepath, positions=None):
    """ Load per cell SNV count data from filepath
    
    Args:
        filepath (str): counts filepath
        positions (pandas.DataFrame): restrict to the specified positions

    Returns:
        pandas.DataFrame: SNV alt and ref counts per cell
    """

    logging.info('Loading snv counts from {}'.format(filepath))

    data = []

    csv_input = scgenome.csvutils.CsvInput(filepath)

    chunk_iter = csv_input.read_csv(
        chunksize=10 ** 6,
        dtypes_override={
            'chrom': 'category',
            'ref': 'category',
            'alt': 'category',
            'cell_id': 'category',
            'sample_id': 'category',
            'library_id': 'category',
        },
    )

    for chunk in chunk_iter:
        if positions is not None:
            scgenome.utils.union_categories(
                [chunk, positions],
                cat_cols=['chrom', 'ref', 'alt'])
            chunk = chunk.merge(positions)

        data.append(chunk)

    data = scgenome.utils.concat_with_categories(data, ignore_index=True)

    logging.info(f'Loaded snv counts table with shape {data.shape}, memory {data.memory_usage().sum()}')

    if positions is not None:
        scgenome.utils.union_categories(
            [data, positions],
            cat_cols=['chrom', 'ref', 'alt'])
        data = data.merge(positions, how='inner')

    logging.info(f'Filtered snv counts table to shape {data.shape}, memory {data.memory_usage().sum()}')

    shape = data.shape
    memory = data.memory_usage().sum()
    logging.info(f'Loaded all snv counts tables with shape {shape}, memory {memory}')

    data['total_counts'] = data['ref_counts'] + data['alt_counts']

    return {'snv_count_data': data}


def load_snv_annotation_table(filepath):
    """ Load SNV annotation data

    Args:
        filepath (str): path of snv annotation table.

    Returns:
        pandas.DataFrame: SNVs annotation data
    """
    logging.info(f'Loading from {filepath}')

    csv_input = scgenome.csvutils.CsvInput(filepath)
    snv_data = csv_input.read_csv(
        dtypes_override={
            'chrom': 'category',
            'ref': 'category',
            'alt': 'category',
            'cell_id': 'category',
            'effect': 'category',
            'effect_impact': 'category',
            'functional_class': 'category',
            'codon_change': 'category',
            'amino_acid_change': 'category',
            'gene_name': 'category',
            'transcript_biotype': 'category',
            'gene_coding': 'category',
            'transcript_id': 'category',
            'genotype': 'category',
        })

    return snv_data


def get_highest_snpeff_effect(snpeff_data):
    """ Select the highest ranked effect from snpeff data.
    """
    ordered_effect_impacts = ['HIGH', 'MODERATE', 'LOW', 'MODIFIER']

    ordered_effect_impacts = pd.DataFrame({
        'effect_impact': ordered_effect_impacts,
        'effect_impact_rank': range(len(ordered_effect_impacts))})
    ordered_effect_impacts['effect_impact'] = (
        ordered_effect_impacts['effect_impact'].astype(snpeff_data['effect_impact'].dtype))

    snpeff_data = snpeff_data.merge(ordered_effect_impacts)
    snpeff_data['coding_rank'] = (snpeff_data['amino_acid_change'].isnull() * 1)

    index_cols = ['chrom', 'coord', 'ref', 'alt']
    value_cols = ['gene_name', 'effect', 'effect_impact', 'amino_acid_change']

    snpeff_data = (snpeff_data[index_cols + value_cols + ['coding_rank', 'effect_impact_rank']]
                   .sort_values(index_cols + ['coding_rank', 'effect_impact_rank'], ascending=True)
                   .groupby(index_cols, sort=False, observed=True)
                   .nth(0)
                   .reset_index()
                   )

    snpeff_data = snpeff_data[[
        'chrom', 'coord', 'ref', 'alt',
        'gene_name', 'effect', 'effect_impact', 'amino_acid_change',
    ]]

    return snpeff_data


def load_snv_annotation_results(
        results_dir,
        museq_filter=None,
        strelka_filter=None,
        mappability_filter=None,
    ):
    """ Load per cell SNV count data from results directory
    
    Args:
        results_dir (str): results directory

    Kwargs:
        museq_filter (float, optional): mutationseq score threshold. Defaults to None.
        strelka_filter (float, optional): strelka score threshold. Defaults to None.
        mappability_filter (float, optional): mappability threshold. Defaults to None.

    Returns:
        pandas.DataFrame: SNV alt and ref per cell
    """

    mappability_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_mappability.csv.gz', analysis_type='variant_calling')

    strelka_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_strelka.csv.gz', analysis_type='variant_calling')

    museq_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_museq.csv.gz', analysis_type='variant_calling')

    cosmic_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_cosmic_status.csv.gz', analysis_type='variant_calling')

    snpeff_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_snpeff.csv.gz', analysis_type='variant_calling')

    dbsnp_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_dbsnp_status.csv.gz', analysis_type='variant_calling')

    trinuc_path = scgenome.loaders.utils.find_results_filepath(
        results_dir, 'snv_trinuc.csv.gz', analysis_type='variant_calling')

    return load_snv_annotation_files(
        mappability_path,
        strelka_path,
        museq_path,
        cosmic_path,
        snpeff_path,
        dbsnp_path,
        trinuc_path,
        museq_filter=museq_filter,
        strelka_filter=strelka_filter,
        mappability_filter=mappability_filter,
    )


def load_snv_annotation_files(
        mappability_path,
        strelka_path,
        museq_path,
        cosmic_path,
        snpeff_path,
        dbsnp_path,
        trinuc_path,
        museq_filter=None,
        strelka_filter=None,
        mappability_filter=None,
    ):
    """ Collate snv results into a single table from input filenames. path inputs must be lists of file strings. 

    Args:
        mappability_path (str): mappability file path
        strelka_path (str): strelka file path
        museq_path (str): museq file path
        cosmic_path (str): cosmic file path
        snpeff_path (str): snpeff file path
        dbsnp_path (str): dbsnp file path
        trinuc_path (str): trinuc file path

    Kwargs:
        museq_filter (float, optional): mutationseq score threshold. Defaults to None.
        strelka_filter (float, optional): strelka score threshold. Defaults to None.
        mappability_filter (float, optional): mappability threshold. Defaults to None.

    """

    if museq_filter is None:
        museq_filter = default_museq_filter

    if strelka_filter is None:
        strelka_filter = default_strelka_filter

    if mappability_filter is None:
        mappability_filter = default_mappability_filter

    logging.info('starting load')

    mappability = load_snv_annotation_table(mappability_path)

    strelka_results = load_snv_annotation_table(strelka_path)
    strelka_results = (
        strelka_results
            .groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['score']
            .max().rename('max_strelka_score').reset_index())

    museq_results = load_snv_annotation_table(museq_path)
    museq_results = (
        museq_results
            .groupby(['chrom', 'coord', 'ref', 'alt'], observed=True)['score']
            .max().rename('max_museq_score').reset_index())

    cosmic = load_snv_annotation_table(cosmic_path)
    logging.info(f'cosmic table with shape {cosmic.shape}, memory {cosmic.memory_usage().sum()}')
    cosmic['is_cosmic'] = 1
    cosmic = cosmic[['chrom', 'coord', 'ref', 'alt', 'is_cosmic']].drop_duplicates()

    snpeff = load_snv_annotation_table(snpeff_path)

    snpeff = get_highest_snpeff_effect(snpeff)
    logging.info(f'snpeff table with shape {snpeff.shape}, memory {snpeff.memory_usage().sum()}')

    dbsnp = load_snv_annotation_table(dbsnp_path)
    dbsnp = dbsnp.query('exact_match == 1')[['chrom', 'coord', 'ref', 'alt']].drop_duplicates().assign(is_dbsnp=1)
    logging.info(f'dbsnp table with shape {dbsnp.shape}, memory {dbsnp.memory_usage().sum()}')

    tnc = load_snv_annotation_table(trinuc_path)

    scgenome.utils.union_categories([
        mappability,
        cosmic,
        snpeff,
        dbsnp,
        tnc,
        strelka_results,
        museq_results,
    ])

    data = pd.concat([
        strelka_results.set_index(['chrom', 'coord', 'ref', 'alt']),
        museq_results.set_index(['chrom', 'coord', 'ref', 'alt']),
    ], axis=1).sort_index().reset_index()
    logging.info(f'merged snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    data = data.merge(mappability, how='left')
    logging.info('post mappability with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    data = data.merge(cosmic, how='left')
    data['is_cosmic'] = data['is_cosmic'].fillna(0).astype(int)
    logging.info('post cosmic with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    data = data.merge(snpeff, how='left')
    logging.info('post snpeff with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    data = data.merge(dbsnp, how='left')
    data['is_dbsnp'] = data['is_dbsnp'].fillna(0).astype(int)
    logging.info('post dbsnp with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    data = data.merge(tnc, how='left')
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    if museq_filter is not None:
        data = data[data['max_museq_score'] >= museq_filter]
        logging.info('post museq filter with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
        logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    if strelka_filter is not None:
        data = data[data['max_strelka_score'] >= strelka_filter]
        logging.info(
            'post strelka filter with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
        logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    if mappability_filter is not None:
        data = data[data['mappability'] >= mappability_filter]
        logging.info(
            'post mappability filter with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
        logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    logging.info('finishing load with snv count {}'.format(data[['chrom', 'coord']].drop_duplicates().shape[0]))
    logging.info(f'snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    for column in categorical_columns:
        data[column] = data[column].astype('category')

    logging.info(f'final snv table with shape {data.shape}, memory {data.memory_usage().sum()}')

    return {'snv_data': data}


def load_snv_results(
        variant_calling_results_dir,
        snv_genotyping_results_dir,
        museq_filter=None,
        strelka_filter=None,
        mappability_filter=None,
    ):
    """ Load filtered SNV annotation and genotyping data
    """

    variant_calling_results_dir = scgenome.loaders.utils.find_results_directory(
        variant_calling_results_dir, 'variant_calling')

    results_tables = load_snv_annotation_results(
        variant_calling_results_dir,
        museq_filter=museq_filter,
        strelka_filter=strelka_filter,
        mappability_filter=mappability_filter,
    )

    snv_genotyping_results_dir = scgenome.loaders.utils.find_results_directory(
        snv_genotyping_results_dir, 'snv_genotyping')

    positions = results_tables['snv_data'][['chrom', 'coord', 'ref', 'alt']].drop_duplicates()

    snv_genotyping_results = load_snv_genotyping_results(snv_genotyping_results_dir, positions)
    results_tables.update(snv_genotyping_results)

    scgenome.utils.union_categories(results_tables.values())

    return results_tables


def load_snv_files(
        mappability_path,
        strelka_path,
        museq_path,
        cosmic_path,
        snpeff_path,
        dbsnp_path,
        trinuc_path,
        counts_path,
        museq_filter=None,
        strelka_filter=None,
        mappability_filter=None,
    ):
    """ Load filtered SNV annotation and genotyping data
    """

    results_tables = load_snv_annotation_files(
        mappability_path,
        strelka_path,
        museq_path,
        cosmic_path,
        snpeff_path,
        dbsnp_path,
        trinuc_path,
        museq_filter=museq_filter,
        strelka_filter=strelka_filter,
        mappability_filter=mappability_filter,
    )

    assert not results_tables['snv_data']['coord'].isnull().any()

    positions = results_tables['snv_data'][['chrom', 'coord', 'ref', 'alt']].drop_duplicates()

    snv_genotyping_results = load_snv_genotyping_files(counts_path, positions)
    results_tables.update(snv_genotyping_results)

    return results_tables

